// Copyright 2013 Yangqing Jia

#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/layer.hpp"
#include "caffe/vision_layers.hpp"
#include "caffe/util/math_functions.hpp"

using std::max;

namespace caffe {

template <typename Dtype>
void GenerativeLossLayer<Dtype>::SetUp(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	CHECK_EQ(bottom.size(), 2) << "GenerativeLoss Layer takes a single blob as input.";
	CHECK_EQ(top->size(), 0) << "GenerativeLoss Layer takes no blob as output.";

	exp_.Reshape(bottom[0]->num(), 1, 1, 1);
	fval_.Reshape(bottom[0]->num(), 1, 1, 1);
	sampleweight_.Reshape(bottom[0]->num(), bottom[0]->num(), 1, 1);
    scale_.Reshape(bottom[0]->num(), 1, 1, 1);
};

template <typename Dtype>
void GenerativeLossLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	const Dtype* bottom_data = bottom[0]->cpu_data();
	const Dtype* label = bottom[1]->cpu_data();
	Dtype* exp_data = exp_.mutable_cpu_data();
	Dtype* fval_data = fval_.mutable_cpu_data();
	Dtype* scale_data = scale_.mutable_cpu_data();
	Dtype* sampleweight_data = sampleweight_.mutable_cpu_data();
	int num = bottom[0]->num();
	int dim = bottom[0]->count() / bottom[0]->num();

	for (int k = 0; k < num; k++) {
		// we need to subtract the max to avoid numerical issues, compute the exp,
		// and then normalize.
		Dtype temp_data = -FLT_MAX;
		for (int i = 0; i < num; ++i) {
		  exp_data[i] = bottom_data[i*dim + static_cast<int>(label[k])];
		  temp_data = max(temp_data, exp_data[i]);
		  //LOG(INFO) << "Generative, label = " << static_cast<int>(label[i]);
		}

		//LOG(INFO) << "label = " << label[0] << " scale_data = " << scale_data[0];

		// subtraction
		for (int i = 0; i < num; ++i) {
		  exp_data[i] -= temp_data;
		}
		fval_data[k] = exp_data[k];

		// Perform exponentiation
		caffe_exp<Dtype>(num, exp_data, exp_data);
		// sum after exp
		scale_data[k] = 0.0;
		for (int i = 0; i < num; ++i) {
			if (i != k) {
				scale_data[k] += (exp_data[i]+FLT_MIN);
			}
		}
		//get sample weight
		for (int i = 0; i < num; ++i) {
			if (i != k) {
				sampleweight_data[k*num+i] = (exp_data[i]+FLT_MIN)/scale_data[k];
			} else {
				sampleweight_data[k*num+i] = 0.0;
			}
		}
		//LOG(INFO) << "sampleweight_data[23] = " << sampleweight_data[23];
	}
}

template <typename Dtype>
void GenerativeLossLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	// TODO(Yangqing): implement the GPU version of softmax.
	return Forward_cpu(bottom, top);
}

template <typename Dtype>
Dtype GenerativeLossLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down,
    vector<Blob<Dtype>*>* bottom) {
	// First, compute the diff
	Dtype* bottom_diff = (*bottom)[0]->mutable_cpu_diff();
	const Dtype* fval_data = fval_.cpu_data();
	const Dtype* scale_data = scale_.cpu_data();
	const Dtype* label = (*bottom)[1]->cpu_data();
	const Dtype* sampleweight_data = sampleweight_.cpu_data();
	int num = (*bottom)[0]->num();
	int dim = (*bottom)[0]->count() / num;

	Dtype loss = 0;
	for (int i = 0; i < (*bottom)[0]->count(); ++i) {
		bottom_diff[i] = 0.0;
	}

	for (int k = 0; k < num; ++k) {
		bottom_diff[k * dim + static_cast<int>(label[k])] -= 1.0;
		for (int i = 0; i < num; ++i) {
			bottom_diff[i * dim + static_cast<int>(label[k])] += sampleweight_data[k*num+i];
		}
		loss += -fval_data[k] + log(max(scale_data[k] / ((Dtype)num-1), FLT_MIN));
	}

	caffe_scal((*bottom)[0]->count(), Dtype(1) / num, bottom_diff);
	return loss/num;
}

template <typename Dtype>
Dtype GenerativeLossLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
  // TODO(Yangqing): implement the GPU version of softmax.
  return Backward_cpu(top, propagate_down, bottom);
}

INSTANTIATE_CLASS(GenerativeLossLayer);


}  // namespace caffe
