// Copyright 2013 Yangqing Jia

#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/layer.hpp"
#include "caffe/vision_layers.hpp"
#include "caffe/util/math_functions.hpp"

using std::max;

namespace caffe {

template <typename Dtype>
void SemisuperGenerativeLossLayer<Dtype>::SetUp(const vector<Blob<Dtype>*>& bottom,
      vector<Blob<Dtype>*>* top) {
	CHECK_EQ(bottom.size(), 2) << "SemisuperGenerativeLoss Layer takes a single blob as input.";
	CHECK_EQ(top->size(), 0) << "SemisuperGenerativeLoss Layer takes no blob as output.";

	int num = bottom[0]->num();
	int dim = bottom[0]->count() / bottom[0]->num();
	fweight_.Reshape(num, dim, 1, 1);
	exp_.Reshape(num, 1, 1, 1);	
	fval_.Reshape(num, dim, 1, 1);
	sampleweight_.Reshape(num, num, dim, 1);
    	scale_.Reshape(num, dim, 1, 1);
};

template <typename Dtype>
void SemisuperGenerativeLossLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	const Dtype* bottom_data = bottom[0]->cpu_data();
	const Dtype* label = bottom[1]->cpu_data();
	Dtype* fweight_data = fweight_.mutable_cpu_data();
	Dtype* exp_data = exp_.mutable_cpu_data();
	Dtype* fval_data = fval_.mutable_cpu_data();
	Dtype* scale_data = scale_.mutable_cpu_data();
	Dtype* sampleweight_data = sampleweight_.mutable_cpu_data();
	int num = bottom[0]->num();
	int dim = bottom[0]->count() / bottom[0]->num();
	Dtype unlabeled_weight = 1;

	for (int k = 0; k < num; k++) {
		Dtype fweight_sum = 0.0;
		for (int l = 0; l < dim; l++) {
			// we need to subtract the max to avoid numerical issues, compute the exp,
			// and then normalize.
			Dtype temp_data = -FLT_MAX;
			for (int i = 0; i < num; ++i) {
			  exp_data[i] = bottom_data[i*dim + l];
			  temp_data = max(temp_data, exp_data[i]);
			}

			// subtraction
			for (int i = 0; i < num; ++i) {
			  exp_data[i] -= temp_data;
			}
			fval_data[k*dim+l] = exp_data[k];

			// Perform exponentiation
			caffe_exp<Dtype>(num, exp_data, exp_data);
			// sum after exp
			scale_data[k*dim+l] = 0.0;
			for (int i = 0; i < num; ++i) {
				scale_data[k*dim+l] += (exp_data[i]+FLT_MIN);
			}
			//get sample weight
			for (int i = 0; i < num; ++i) {
				sampleweight_data[k*num*dim+i*dim+l] = (exp_data[i]+FLT_MIN)/scale_data[k*dim+l];
			}
			
			if (static_cast<int>(label[k]) == -1) {
				fweight_data[k*dim+l] = exp(fval_data[k*dim+l] - log(max(scale_data[k*dim+l] / ((Dtype)num), FLT_MIN)));
				// LOG(INFO) << "fweight_data = " << fweight_data[k*dim+l];	
			}
			else {
				fweight_data[k*dim+l] = static_cast<int>(label[k])==l?1:0;	
			}
			fweight_sum += 	fweight_data[k*dim+l];
		}
		if (static_cast<int>(label[k]) == -1) {
			for (int l = 0; l < dim; l++) {
				fweight_data[k*dim+l] = fweight_data[k*dim+l] * unlabeled_weight / fweight_sum;
			}
		}
		
	}
	
}

template <typename Dtype>
void SemisuperGenerativeLossLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
    vector<Blob<Dtype>*>* top) {
	// TODO(Yangqing): implement the GPU version of softmax.
	return Forward_cpu(bottom, top);
}

template <typename Dtype>
Dtype SemisuperGenerativeLossLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down,
    vector<Blob<Dtype>*>* bottom) {
	// First, compute the diff
	Dtype* bottom_diff = (*bottom)[0]->mutable_cpu_diff();
	const Dtype* fweight_data = fweight_.cpu_data();
	const Dtype* fval_data = fval_.cpu_data();
	const Dtype* scale_data = scale_.cpu_data();
	const Dtype* sampleweight_data = sampleweight_.cpu_data();
	const Dtype* label = (*bottom)[1]->cpu_data();
	int num = (*bottom)[0]->num();
	int dim = (*bottom)[0]->count() / num;

	
	for (int i = 0; i < (*bottom)[0]->count(); ++i) {
		bottom_diff[i] = 0.0;
	}

	Dtype loss = 0;
	for (int k = 0; k < num; ++k) {
		Dtype cur_loss = 0;
		for (int l = 0; l < dim; l++) {
			bottom_diff[k * dim + l] -= 1.0 * fweight_data[k*dim+l];
			for (int i = 0; i < num; ++i) {
				bottom_diff[i * dim + l] += sampleweight_data[k*num*dim+i*dim+l] * fweight_data[k*dim+l];
			}
			if (static_cast<int>(label[k]) == -1) {
				cur_loss += exp(fval_data[k*dim+l] - log(max(scale_data[k*dim+l] / ((Dtype)num), FLT_MIN)));
			}
		}
		if (static_cast<int>(label[k]) == -1) {
			loss += -log(cur_loss);
		}
		else {
			loss += -fval_data[k*dim+static_cast<int>(label[k])] + log(max(scale_data[k*dim+static_cast<int>(label[k])] / ((Dtype)num), FLT_MIN));
			//LOG(INFO) << "loss = " << loss;
		}
	}
	// LOG(INFO) << "Entropy = " << neg_entropy/num;

	caffe_scal((*bottom)[0]->count(), Dtype(1) / num, bottom_diff);
	return loss/num;
}

template <typename Dtype>
Dtype SemisuperGenerativeLossLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
    const bool propagate_down, vector<Blob<Dtype>*>* bottom) {
  // TODO(Yangqing): implement the GPU version of softmax.
  return Backward_cpu(top, propagate_down, bottom);
}

INSTANTIATE_CLASS(SemisuperGenerativeLossLayer);


}  // namespace caffe
